// Nodes : input.cpp
//
// Read various types of input files
//
// See userguide.pdf for copyright information

#include <fstream>
#include <string>
#include <map>

using namespace std;

#include <math.h>

#include "realnum.h"
#include "universe.h"
#include "node.h"
#include "input.h"



CNode::NodeType GetNodeType(char *s)
{
  if (string(s) == "ins")
    return CNode::null;
  else if (string(s) == "wire")
    return CNode::wire;
  else if (string(s) == "not")
    return CNode::not;
  else if (string(s) == "thrust")
    return CNode::thrust;
  else if (string(s) == "fuse")
    return CNode::fuse;
  else if (string(s) == "rfuse")
    return CNode::rfuse;
  else if (string(s) == "lfuse")
    return CNode::lfuse;
  else if (string(s) == "toggle")
    return CNode::toggle;
  else if (string(s) == "delta")
    return CNode::delta;
  else if (string(s) == "ndelta")
    return CNode::ndelta;
  else if (string(s) == "detect")
    return CNode::detect;
  else if (string(s) == "or")
    return CNode::or;
  else if (string(s) == "runfuse")
    return CNode::runfuse;
  else if (string(s) == "lunfuse")
    return CNode::lunfuse;
  else if (string(s) == "cross")
    return CNode::cross;
  else if (string(s) == "keypress")
    return CNode::keypress;
  else if (string(s) == "pulse")
    return CNode::pulse;
  else if (string(s) == "rslide")
    return CNode::rslide;
  else if (string(s) == "lslide")
    return CNode::lslide;
  else if (string(s) == "store")
    return CNode::store;
  else if (string(s) == "creator")
    return CNode::creator;
  else if (string(s) == "maj")
    return CNode::maj;
  else if (string(s) == "notnot")
    return CNode::notnot;
  else if (string(s) == "push")
    return CNode::push;
  else if (string(s) == "equal")
    return CNode::equal;
  else
    return CNode::null;
}

void ReadInput(const char *fname, CUniverse *U)
{
  ifstream is(fname);

  map<string,CNode *> varmap;
  char token[128];
  char id[128];

  while(is)
  {
    is >> token;

    if (string(token) == "node")
    {
      RealNum x,y,a,r;
      CNode *n;

      is >> id;
      is >> token;
      is >> x >> y >> a >> r;

      while(a>=2.0*PI)
        a-=2.0*PI;

      while(a<0)
        a+=2.0*PI;

      n = new CNode(x,y,a,0,0,0,r,GetNodeType(token));

      U->AddNode(n);
      varmap[string(id)] = n;
    }
    else if (string(token) == "nodeatpos")
    {
      RealNum x,y;
      CNode *n;

      is >> id;
      is >> x >> y;

      n = U->FindNode(x,y,8,NULL);
      varmap[string(id)] = n;
    }
    else if (string(token) == "connect")
    {
      is >> id;
      is >> token;

      if ( varmap.find(string(id)) != varmap.end() &&
           varmap.find(string(token)) != varmap.end())
      {
        varmap[string(id)]->Connect(varmap[string(token)]);
      }
    }
    else if (string(token) == "setnum")
    {
      int num;

      is >> id >> num;

      if (varmap.find(string(id)) != varmap.end())
      {
        varmap[string(id)]->SetNum(num);
      }
    }
  }
}

CNode::CNode(ifstream &is, unsigned &label)
{
  int tp;
  unsigned connode;

  is >> tp;

  if (!is)
  {
    label = 0;
    connections[0].node = connections[1].node = connections[2].node = connections[3].node = NULL;
    return;
  }

  type = (NodeType)tp;
  is >> xpos >> ypos >> angle;
  is >> xdelta >> ydelta >> angledelta >> radius;
  is >> output[0] >> output[1] >> output[2] >> output[3];
  is >> state;
  is >> label;

  sa = sin(angle);
  ca = cos(angle);

  for(unsigned i = 0; i<4; i++)
  {
    is >> connections[i].dir;
    is >> connode;
    connections[i].node = (CNode *)connode;
  }

  buffer[0] = output[0];
  buffer[1] = output[1];
  buffer[2] = output[2];
  buffer[3] = output[3];

  moveiteration = 0;
}

void ReadDumpInput(const char *fname,CUniverse *U)
{
  ifstream is(fname);

  map<unsigned,CNode *> varmap;
  unsigned label;

  varmap[0] = NULL;

  while(is)
  {
    CNode *n = new CNode(is,label);

    if (!label)
    {
      delete n;
    }
    else
    {
      varmap[label] = n;

      U->AddNode(n);
    }
  }

  for(map<unsigned,CNode *>::iterator i=varmap.begin(); i!=varmap.end(); i++)
  {
	if ((*i).first != 0)
      for(unsigned j=0; j<4; j++)
	  {
        (*i).second->GetConnection(j)->node =
          varmap[(unsigned)((*i).second->GetConnection(j)->node)];
	  }
  }
}

void ReadChangeInput(const char *fname, CUniverse *U)
{
  ifstream is(fname);
  char token[32];

  while(is)
  {
    is >> token;

    if (!stricmp(token,"rep"))
    {
      RealNum x,y;
      unsigned dir;
      is >> x >> y >> token >> dir;

      CNode *n = U->FindNode(x,y,8,NULL);

      if(n)
      {
        n->Change(GetNodeType(token),dir);
      }
    }
    else if (!stricmp(token,"exc"))
    {
      RealNum x1,y1,x2,y2;
      is >> x1 >> y1 >> x2 >> y2;

      U->Exclude(x1,y1,x2,y2);
    }
  }
}

