// Nodes : node.cpp
//
// Deal with motion, interaction, connection and function of nodes
//
// See userguide.pdf for copyright information

#include <math.h>
#include <assert.h>

#include "realnum.h"
#include "node.h"
#include "universe.h"
#include "constants.h"

int dircontribs[4][2][2] = 
{
	{{1,0},{0,1}},
	{{0,1},{-1,0}},
	{{-1,0},{0,-1}},
	{{0,-1},{1,0}},
};

#define SINADIR(SA,CA,D) (dircontribs[D][0][0]*SA+dircontribs[D][0][1]*CA)
#define COSADIR(SA,CA,D) (dircontribs[D][1][0]*SA+dircontribs[D][1][1]*CA)

CNode::CNode(RealNum x, RealNum y, RealNum a, RealNum xd, RealNum yd, RealNum ad, RealNum r, NodeType nt)
  :  type(nt), xpos(x), ypos(y), angle(a), xdelta(xd), ydelta(yd), angledelta(ad), radius(r)
{
  if (nt>=numtypes)
  {
    assert(0);
  }

  while (angle >= 2.0*PI)
    angle = angle - 2.0*PI;

  while (angle < 0)
    angle = angle + 2.0*PI;

  for(unsigned i = 0; i<4; i++)
  {
    connections[i].node = NULL;
    output[i] = 0;
    buffer[i] = 0;
  }

  sa = sin(angle);
  ca = cos(angle);

  state = 0;
  moveiteration = 0;
}

CNode::~CNode(void)
{
  for(unsigned i = 0; i<4; i++)
  {
    if (connections[i].node)
      connections[i].node->connections[connections[i].dir].node = NULL;
  }
}

bool CNode::Move(unsigned iteration)
{
  if (moveiteration != iteration)
  {
    xpos += xdelta;
    ypos += ydelta;
    angle += angledelta;

    if (angle >= 2.0*PI)
      angle = angle - 2.0*PI;

    if (angle < 0)
      angle = angle + 2.0*PI;

	sa = sin(angle);
	ca = cos(angle);

    moveiteration++;
    return true;
  }
  else
  {
    return false;
  }
}

void CNode::Forces(CNode *n)
{
  RealNum dx = n->xpos - xpos;
  RealNum dy = n->ypos - ypos;

  RealNum dist2 = dx*dx+dy*dy;
  RealNum dist = sqrt(dist2);

  if (dist < radius + n->radius)
  {
    RealNum f = (radius + n->radius) - dist;

    f = f*f * REPEL_FORCE_CONSTANT / dist;

    xdelta -= f * dx;
    ydelta -= f * dy;

    n->xdelta += f * dx;
    n->ydelta += f * dy;
  }
}

void CNode::ConnectionForces(void)
{
  for(unsigned i = 0; i<4; i++)
  {
    if (connections[i].node)
    {
      RealNum rqdx = connections[i].node->xpos +
        (radius+connections[i].node->radius)*SINADIR(connections[i].node->sa,connections[i].node->ca,connections[i].dir);
      RealNum rqdy = connections[i].node->ypos +
        (radius+connections[i].node->radius)*COSADIR(connections[i].node->sa,connections[i].node->ca,connections[i].dir);

      RealNum fx = CONNECT_FORCE_CONSTANT * (rqdx - xpos);
      RealNum fy = CONNECT_FORCE_CONSTANT * (rqdy - ypos);

      xdelta+=fx;
      ydelta+=fy;

      connections[i].node->xdelta -= fx;
      connections[i].node->ydelta -= fy;

      RealNum rqdtheta,adiff;

#if 1
      RealNum dx = connections[i].node->xpos - xpos;
      RealNum dy = connections[i].node->ypos - ypos;

      RealNum a = atan2(dy,dx);
      a = PI/2.0 - a;
      if (a<0) a += 2.0*PI;

      rqdtheta = - (RealNum)i*PI/2.0 + a;
      if (rqdtheta<0) rqdtheta += 2.0*PI;
      
      if (angle > rqdtheta)
      {
        adiff = angle - rqdtheta;

        if (adiff > PI)
        {
          adiff = adiff-2.0*PI;
        }

        adiff = -adiff;
      }
      else
      {
        adiff = rqdtheta - angle;

        if (adiff > PI)
        {
          adiff = adiff-2.0*PI;
        }
      }

      angledelta += ANGLE_FORCE_CONSTANT*adiff;
#endif

      rqdtheta = connections[i].node->angle + (RealNum)connections[i].dir * PI/2.0
        + PI - (RealNum)i * PI/2.0;

      while(rqdtheta >= 2.0*PI)
        rqdtheta -= 2.0*PI;

      while(rqdtheta < 0)
        rqdtheta += 2.0*PI;

      if (angle > rqdtheta)
      {
        adiff = angle - rqdtheta;

        if (adiff > PI)
        {
          adiff = adiff-2.0*PI;
        }

        adiff = -adiff;
      }
      else
      {
        adiff = rqdtheta - angle;

        if (adiff > PI)
        {
          adiff = adiff-2.0*PI;
        }
      }

      angledelta += SAME_ANGLE_FORCE_CONSTANT * adiff;

    }
  }
}

void CNode::FrictionForces(void)
{
  angledelta *= ANGLE_FRICTION_FORCE_CONSTANT;
  xdelta *= FRICTION_FORCE_CONSTANT;
  ydelta *= FRICTION_FORCE_CONSTANT;
}

void CNode::Connect(CNode *n)
{
  if (n==this)
    return;

  // Find the angle between n and this node
  RealNum dx = n->xpos - xpos;
  RealNum dy = n->ypos - ypos;

  RealNum a = atan2(dy,dx);
  a = PI/2.0 - a;
  if (a<0) a += 2.0*PI;

  unsigned con = ((unsigned)((a + PI/4.0 - angle + 2.0*PI)/(PI/2.0)))%4;
  unsigned conn = ((unsigned)((a + PI/4.0 - n->angle + 3.0*PI)/(PI/2.0)))%4;

  if (connections[con].node == NULL && n->connections[conn].node == NULL)
  {
    connections[con].node = n;
    connections[con].dir = conn;
    n->connections[conn].node = this;
    n->connections[conn].dir = con;
  }
}

void CNode::Disconnect(CNode *n)
{
  if (n==this)
    return;

  // Find the angle between n and this node
  RealNum dx = n->xpos - xpos;
  RealNum dy = n->ypos - ypos;

  RealNum a = atan2(dy,dx);
  a = PI/2.0 - a;
  if (a<0) a += 2.0*PI;

  unsigned con = ((unsigned)((a + PI/4.0 - angle + 2.0*PI)/(PI/2.0)))%4;
  unsigned conn = ((unsigned)((a + PI/4.0 - n->angle + 3.0*PI)/(PI/2.0)))%4;

  if (connections[con].node == n && n->connections[conn].node == this)
  {
    connections[con].node = NULL;
    n->connections[conn].node = NULL;
  }
}

void CNode::SetNum(int n)
{
  buffer[0] = n;
}

void CNode::Process(CUniverse *u, CUniverseCell *uc)
{
  int tmp;

  switch(type)
  {
    case null:
      break;
    case wire:
      buffer[0] = GetInput(2,u,uc);
      break;
    case not:
      buffer[0] = !GetInput(2,u,uc);
      break;
    case notnot:
      buffer[0] = !!GetInput(2,u,uc);
      break;
    case thrust:
      if (GetInput(2,u,uc))
      {
        ApplyForce(-THRUST_FORCE_CONSTANT,angle);
      }
      break;
    case fuse:
      if (GetInput(2,u,uc))
      {
        CNode *n1 =
        u->FindNode(xpos + 2*radius*sa,
                    ypos + 2*radius*ca,
                    1.5*radius,this);
        CNode *n2 =
        u->FindNode(xpos + 2.828*radius*sin(angle+PI/4.0),
                    ypos + 2.828*radius*cos(angle+PI/4.0),
                    1.5*radius,this);
        CNode *n3 =
        u->FindNode(xpos + 2.828*radius*sin(angle-PI/4.0),
                    ypos + 2.828*radius*cos(angle-PI/4.0),
                    1.5*radius,this);

        if (n1 && n2)
          n1->Connect(n2);
        if (n1 && n3)
          n1->Connect(n3);
      }
      break;
    case rfuse:
      if (GetInput(2,u,uc))
      {
        CNode *n1 =
        u->FindNode(xpos + 2*radius*sa,
                    ypos + 2*radius*ca,
                    1.5*radius,this);
        CNode *n2 =
        u->FindNode(xpos + 2.828*radius*sin(angle+PI/4.0),
                    ypos + 2.828*radius*cos(angle+PI/4.0),
                    1.5*radius,this);
        if (n1 && n2)
          n1->Connect(n2);
      }
      break;
    case lfuse:
      if (GetInput(2,u,uc))
      {
        CNode *n1 =
        u->FindNode(xpos + 2*radius*sa,
                    ypos + 2*radius*ca,
                    1.5*radius,this);
        CNode *n2 =
        u->FindNode(xpos + 2.828*radius*sin(angle-PI/4.0),
                    ypos + 2.828*radius*cos(angle-PI/4.0),
                    1.5*radius,this);

        if (n1 && n2)
          n1->Connect(n2);
      }
      break;
    case toggle:
      {
        int i1 = GetInput(1,u,uc);
        int i3 = GetInput(3,u,uc);

        if (i1 && !i3)
        {
          buffer[0] = buffer[2] = 1;
        }
        else if (i3 && !i1)
        {
          buffer[0] = buffer[2] = 0;
        }
      }
      break;
    case delta:
      buffer[0] = buffer[1] = buffer[3] = GetInput(2,u,uc);
      break;
    case ndelta:
      buffer[0] = buffer[1] = buffer[3] = !GetInput(2,u,uc);
      break;
    case detect:
      {
        CNode *n =
        u->FindNode(xpos + 2*radius*sa,
                    ypos + 2*radius*ca,
                    radius,this);

        if (n)
        {
          RealNum a = n->angle+PI/4.0;

          while(a<0)
            a+=2.0*PI;

          while(a>=2.0*PI)
            a-=2.0*PI;

          buffer[2] = 1+(int)(a/(PI/2.0)) + ((int)(n->type))*4;
        }
        else
        {
          buffer[2] = 0;
        }
      }
      break;
    case or:
      buffer[0] = GetInput(1,u,uc) + GetInput(3,u,uc);
      break;
    case runfuse:
      if (GetInput(2,u,uc))
      {
        if (connections[0].node)
        {
          CNodeConnection *dis = &connections[0].node->connections[(connections[0].dir+3)%4];

          if (dis->node)
          {
            dis->node->connections[dis->dir].node = NULL;
            dis->node = NULL;
          }
        }
        else
        {
          CNode *n1 =
          u->FindNode(xpos + 2*radius*sa,
                    ypos + 2*radius*ca,
                    1.5*radius,this);
          CNode *n2 =
          u->FindNode(xpos + 2.828*radius*sin(angle+PI/4.0),
                    ypos + 2.828*radius*cos(angle+PI/4.0),
                    1.5*radius,this);
          if (n1 && n2)
            n1->Disconnect(n2);
        }
      }
      break;
    case lunfuse:
      if (GetInput(2,u,uc))
      {
        if (connections[0].node)
        {
          CNodeConnection *dis = &connections[0].node->connections[(connections[0].dir+1)%4];

          if (dis->node)
          {
            dis->node->connections[dis->dir].node = NULL;
            dis->node = NULL;
          }
        }
        else
        {
          CNode *n1 =
          u->FindNode(xpos + 2*radius*sa,
                    ypos + 2*radius*ca,
                    1.5*radius,this);
          CNode *n2 =
          u->FindNode(xpos + 2.828*radius*sin(angle-PI/4.0),
                    ypos + 2.828*radius*cos(angle-PI/4.0),
                    1.5*radius,this);
          if (n1 && n2)
            n1->Disconnect(n2);
        }
      }
      break;
    case cross:
      buffer[0] = GetInput(2,u,uc);
      buffer[1] = GetInput(3,u,uc);
      break;
    case keypress:
      switch(GetInput(2,u,uc))
      {
#if 0
		case 1:
          buffer[0] = !!key[KEY_1];
          break;
        case 2:
          buffer[0] = !!key[KEY_2];
          break;
        case 3:
          buffer[0] = !!key[KEY_3];
          break;
        case 4:
          buffer[0] = !!key[KEY_4];
          break;
        case 5:
          buffer[0] = !!key[KEY_5];
          break;
        case 6:
          buffer[0] = !!key[KEY_6];
          break;
        case 7:
          buffer[0] = !!key[KEY_7];
          break;
        case 8:
          buffer[0] = !!key[KEY_8];
          break;
        case 9:
          buffer[0] = !!key[KEY_9];
          break;
#endif
        default:
          buffer[0] = 0;
          break;
      }
      break;
    case pulse:
      {
        int i;

        i=GetInput(2,u,uc);

        if (i)
        {
          if (state != i)
          {
            buffer[0] = i;
            state = i;
          }
          else
          {
            buffer[0] = 0;
          }
        }
        else
        {
          state = 0;
          buffer[0] = 0;
        }
      }
      break;
    case rslide:
      if (GetInput(2,u,uc))
      {
        CNode *n =
        u->FindNode(xpos + 2*radius*sa,
                    ypos + 2*radius*ca,
                    radius,this);

        if (n)
        {
          n->ApplyForce(THRUST_FORCE_CONSTANT,angle+PI/2.0);
        }
      }
      break;
    case lslide:
      if (GetInput(2,u,uc))
      {
        CNode *n =
        u->FindNode(xpos + 2*radius*sa,
                    ypos + 2*radius*ca,
                    radius,this);

        if (n)
        {
          n->ApplyForce(THRUST_FORCE_CONSTANT,angle-PI/2.0);
        }
      }
      break;
    case store:
      if (buffer[0]==0 && (tmp=GetInput(2,u,uc)))
      {
        buffer[0] = tmp;
      }
      else if (GetInput(1,u,uc) || GetInput(3,u,uc))
      {
        buffer[0] = 0;
      }
      break;
    case creator:
      if ((tmp=GetInput(2,u,uc)) && tmp<=numtypes*4)
      {
        RealNum nx = xpos + 2*radius*sa;
        RealNum ny = ypos + 2*radius*ca;

        CNode *n =
          u->FindNode(nx,ny,2*radius,this);

        if (!n)
        {
          n = new CNode(nx,ny,angle+((tmp-1)%4)*PI/2.0,0.0,0.0,0.0,radius,(NodeType)(((tmp-1)/4)%numtypes));
          if (n)
            u->CreateNode(n);
          else
          {
            assert(0);
          }
        }
      }
      break;
    case maj:
      {
        int i1 = GetInput(1,u,uc);
        int i2 = GetInput(2,u,uc);
        int i3 = GetInput(3,u,uc);

        buffer[0] = (i1 * i3) + (i1 * i2) + (i2 * i3);
      }
      break;
    case push:
      if (GetInput(2,u,uc))
      {
        CNode *n =
        u->FindNode(xpos + 2*radius*sa,
                    ypos + 2*radius*ca,
                    radius,this);

        if (n)
        {
          n->ApplyForce(THRUST_FORCE_CONSTANT,angle);
        }
      }
      break;
    case equal:
      {
        int i1 = GetInput(1,u,uc);
        int i2 = GetInput(2,u,uc);
        int i3 = GetInput(3,u,uc);

        buffer[0] = ((i1 == i3) && i1) ||
                  ((i1 == i2) && i1) ||
                  ((i2 == i3) && i2);
      }
      break;
    default:
      assert(0);
      break;
  }
}

void CNode::Update(void)
{
  for(unsigned i = 0; i<4; i++)
  {
    output[i] = buffer[i];
  }
}

RealNum CNode::DistanceSquared(RealNum x, RealNum y)
{
  return (x-xpos)*(x-xpos) + (y-ypos)*(y-ypos);
}

int CNode::GetInput(unsigned d, CUniverse *u, CUniverseCell *uc)
{
  if (connections[d].node)
  {
    return connections[d].node->output[connections[d].dir];
  }
  else
  {
    return uc->GetSignalValue(
      xpos + radius * SINADIR(sa,ca,d),
      ypos + radius * COSADIR(sa,ca,d),
      radius/2,
      this);
  }
}

void CNode::GetSignalValue(RealNum x, RealNum y, RealNum &r2, int &rval)
{
  for(unsigned i = 0; i<4; i++)
  {
    RealNum xp = xpos + radius*SINADIR(sa,ca,i);
    RealNum yp = ypos + radius*COSADIR(sa,ca,i);

    RealNum dist2 = (x-xp)*(x-xp) + (y-yp)*(y-yp);

    if (dist2 <= r2)
    {
      rval = output[i];
      r2 = dist2;
    }
  }
}

void CNode::ApplyForce(RealNum f, RealNum a)
{
  xdelta += f * sin(a);
  ydelta += f * cos(a);
}

/* TODO - customize this function for your graphics routines.
 * 'param' may be used to pass parameters for your graphics routines,
 * 'options' may be used to specify drawing options.
 * define the shape macros to suite your purposes.
 *
 * The value of 'scale' may depend upon information passed through 'param'
 */
void CNode::Draw(void *param,RealNum x, RealNum y, RealNum w, RealNum h, unsigned options)
{
  int centrepos[2];
  int dirpos[4][2];

  float scale = 1;

  centrepos[0] = (int) ((xpos-x)*scale);
  centrepos[1] = (int) ((ypos-y)*scale);


  {for(unsigned i = 0; i<4; i++)
  {
    RealNum a = angle + (RealNum)i * PI/2.0;

    RealNum xo = radius * sin(a);
    RealNum yo = radius * cos(a);

    dirpos[i][0] = (int) (xo*scale);
    dirpos[i][1] = (int) (yo*scale);
  }}

#define LINE(x1,y1,x2,y2)

#define CIRCLE(x1,y1,rad)

#define CIRCLEFILL(x1,y1,rad) 

#define CIRCLEFILLSIGNAL(x1,y1,rad)

#define CIRCLEERASE(x1,y1,rad) 
#define TARGETCIRCLE(x1,y1,rad)

#include "glyphs.h"

  if (1)
  {
	{for(unsigned i = 0; i<4; i++)
	{
		RealNum a = (RealNum)i * PI/2.0;

		RealNum xo = sin(a);
		RealNum yo = cos(a);

		if (output[i])
		{
			CIRCLEFILLSIGNAL(xo,yo,0.125);
		}
		else
		{
			CIRCLE(xo,yo,0.125);
		}

                if (output[i])
		{
			char numstring[16];

			sprintf(numstring,"%d",output[i]);

			int x =	(int)(centrepos[0]+xo*dirpos[1][0]+yo*dirpos[0][0]);
			int y =	(int)(centrepos[1]+xo*dirpos[1][1]+yo*dirpos[0][1]);

			CFont f;
			f.CreatePointFont(scale*30,"Arial",pDC);

			CFont *oldfont = pDC->SelectObject(&f);
			pDC->SetTextColor(RGB(20,150,20));
			int oldbkmode = pDC->SetBkMode(TRANSPARENT);
			pDC->TextOut(x,y,numstring);
			pDC->SelectObject(oldfont);
			pDC->SetBkMode(oldbkmode);

		}
	}}
  }
}
#undef LINE
#undef CIRCLE
#undef CIRCLEFILL
#undef CIRCLEFILLSIGNAL
#undef CIRCLEERASE
#undef TARGETCIRCLE

static void PostscriptLine(ofstream &os,RealNum x1,RealNum y1,RealNum x2,RealNum y2)
{
  os << x1 << " " << y1 << " moveto" << endl;
  os << x2 << " " << y2 << " lineto" << endl;
  os << "stroke" << endl;
}

static void PostscriptCircle(ofstream &os,RealNum x1,RealNum y1,RealNum rad)
{
  float xs,ys;

  if (rad<1)
    rad = 1;

  os << x1 << " " << y1+rad << " moveto" << endl;

  for(float a = 0; a<2*PI; a+=0.1*PI)
  {
    xs = x1 + rad*sin(a);
    ys = y1 + rad*cos(a);

    os << xs << " " << ys << " lineto" << endl;
  }

  xs = x1 + rad*sin(a);
  ys = y1 + rad*cos(a);
  os << xs << " " << ys << " lineto" << endl;
}

static void PostscriptCircleFill(ofstream &os,RealNum x1,RealNum y1,RealNum rad)
{
  PostscriptCircle(os,x1,y1,rad);
  os << "fill" << endl;
  os << "stroke" << endl;
}

void CNode::DrawPostscript(ofstream &os,RealNum x, RealNum y, RealNum w, RealNum h)
{
  if (xpos < (x-3*radius) || xpos > (x+w+3*radius) ||
      ypos < (y-3*radius) || ypos > (y+h+3*radius))
  {
    return;
  }

  int centrepos[2];
  int dirpos[4][2];

  RealNum scale = 4000/w;

  centrepos[0] = (int) ((xpos-x)*scale);
  centrepos[1] = (int) ((ypos-y)*scale);

  {for(unsigned i = 0; i<4; i++)
  {
    RealNum a = angle + (RealNum)i * PI/2.0;

    RealNum xo = radius * sin(a);
    RealNum yo = radius * cos(a);

    dirpos[i][0] = (int) (xo*scale);
    dirpos[i][1] = (int) (yo*scale);
  }}

#define LINE(x1,y1,x2,y2) \
  PostscriptLine(os, \
    (int)(centrepos[0]+x1*dirpos[1][0]+y1*dirpos[0][0]), \
    (int)(centrepos[1]+x1*dirpos[1][1]+y1*dirpos[0][1]), \
    (int)(centrepos[0]+x2*dirpos[1][0]+y2*dirpos[0][0]), \
    (int)(centrepos[1]+x2*dirpos[1][1]+y2*dirpos[0][1]));

#define CIRCLE(x1,y1,rad) \
  PostscriptCircle(os, \
    (int)(centrepos[0]+x1*dirpos[1][0]+y1*dirpos[0][0]), \
    (int)(centrepos[1]+x1*dirpos[1][1]+y1*dirpos[0][1]), \
    (int)(rad * scale * radius)); \
  os << "stroke" << endl;

#define CIRCLEFILL(x1,y1,rad) \
  PostscriptCircleFill(os, \
    (int)(centrepos[0]+x1*dirpos[1][0]+y1*dirpos[0][0]), \
    (int)(centrepos[1]+x1*dirpos[1][1]+y1*dirpos[0][1]), \
    (int)(rad * scale * radius));

#define CIRCLEERASE(x1,y1,rad)

#define TARGETCIRCLE(x1,y1,rad)

#include "glyphs.h"

  {for(unsigned i = 0; i<4; i++)
  {
    RealNum xo = ( (i==1)?1.0:( (i==3)?-1.0:0.0 ) );
    RealNum yo = ( (i==0)?1.0:( (i==2)?-1.0:0.0 ) );

    if (output[i])
    {
      CIRCLEFILL(xo,yo,0.125);
    }
    else
    {
      CIRCLE(xo,yo,0.125);
    }
  }}

}

void CNode::DrawAnim(ofstream &os)
{
  os << (int)type << " " << (int)xpos << " " << (int)ypos << endl;
}

void CNode::Change(NodeType nt, unsigned dir)
{
  if (nt != type)
  {
    for(unsigned i = 0; i<4; i++)
    {
      output[i] = buffer[i] = 0;
    }
  }

  type = nt;

  RealNum anglediff;
  CNodeConnection tempc;

  while(1)
  {
    anglediff = angle - (dir*PI/2);
    if (anglediff>PI) anglediff = anglediff - 2*PI;
    if (anglediff<-PI) anglediff = anglediff + 2*PI;

    if (anglediff>PI/4)
    {
      angle = angle - PI/2;

      tempc = connections[3];
      for(unsigned i = 3; i>0; i--)
      {
        connections[i] = connections[i-1];
        if (connections[i].node)
          connections[i].node->connections[connections[i].dir].dir = i;
      }
      connections[0] = tempc;
      if (connections[0].node)
        connections[0].node->connections[connections[0].dir].dir = 0;
    }
    else if (anglediff<-PI/4)
    {
      angle = angle + PI/2;

      tempc = connections[0];
      for(unsigned i = 0; i<3; i++)
      {
        connections[i] = connections[i+1];
        if (connections[i].node)
          connections[i].node->connections[connections[i].dir].dir = i;
      }
      connections[3] = tempc;
      if (connections[3].node)
        connections[3].node->connections[connections[3].dir].dir = 3;
    }
    else
      break;
  }

  while (angle >= 2.0*PI)
    angle = angle - 2.0*PI;

  while (angle < 0)
    angle = angle + 2.0*PI;

  sa = sin(angle);
  ca = cos(angle);
}
