// Nodes : Universe.cpp
//
// Implement a universe divided up into small squares (cells).
// The universe expands and contracts itself as necessary.
//
// See userguide.pdf for copyright information

#include <map>
#include <list>

using namespace std;

#include <math.h>
#include <assert.h>

#include "realnum.h"
#include "universe.h"

#define NEIGHBOUR(dx,dy) (((dx)+1)*3+((dy)+1))

#define POSTOCELLX(x,cx) (((int)((x)/(cx)))-((x)<0))
#define POSTOCELLY(y,cy) (((int)((y)/(cy)))-((y)<0))

RealNum CUniverseCell::width = 32;
RealNum CUniverseCell::height = 32;

CUniverseCell::CUniverseCell(void)
{
  numnodes = 0;
  accessed = true;

  for(unsigned i = 0; i<9; i++)
  {
    neighbours[i] = NULL;
  }
}

void CUniverseCell::SetPosition(RealNum x, RealNum y)
{
  xpos = x;
  ypos = y;
}

void CUniverseCell::SetNeighbour(unsigned n, CUniverseCell *uc)
{
  neighbours[n] = uc;
}

void CUniverseCell::AddNode(CNode *n)
{
  assert(numnodes < MAX_NODES_PER_CELL);
  nodes[numnodes++] = n;
  accessed = true;
}

void CUniverseCell::Move(CUniverse *u)
{
  int dx,dy;
  unsigned n;
  RealNum nx,ny;
  unsigned it = u->GetIteration();

  for(CNode **i = &nodes[0]; i<&(nodes[0])+numnodes;)
  {
    if ((*i)->Move(it))
    {
      nx = (*i)->GetXpos();
      ny = (*i)->GetYpos();

      dx = ( nx >= xpos+width );
      dx -= ( nx < xpos );

      dy = ( ny >= ypos+height );
      dy -= ( ny < ypos );

      if (dx || dy)
      {
        n = NEIGHBOUR(dx,dy);

        if (!neighbours[n])
        {
          u->AddCell(xpos+width/2+dx*width,ypos+height/2+dy*height);
        }

        neighbours[n]->AddNode(*i);

        *i = nodes[--numnodes];
      }
      else
      {
        i++;
      }
    }
    else
    {
      i++;
    }
  }

  if (numnodes) accessed = true;
}

void CUniverseCell::Forces(void)
{
  for(CNode **i = &nodes[0]; i<&(nodes[0])+numnodes; i++)
  {
    for(unsigned n = NEIGHBOUR(0,0); n<=NEIGHBOUR(1,1); n++)
    {
      if (neighbours[n])
      {
        if (neighbours[n] != this)
        {
          for(CNode **j = &(neighbours[n]->nodes[0]); j < &(neighbours[n]->nodes[0])+neighbours[n]->numnodes; j++)
          {
            (*i)->Forces(*j);
          }
        }
        else
        {
          for(CNode **j = i; j<&(nodes[0])+numnodes; j++)
          {
            if (*i != *j)
              (*i)->Forces(*j);
          }
        }
      }
    }

    (*i)->ConnectionForces();
  }
}

void CUniverseCell::FrictionForces(void)
{
  for(CNode **i = &nodes[0]; i<&(nodes[0])+numnodes; i++)
    (*i)->FrictionForces();
}

void CUniverseCell::Process(CUniverse *u)
{
  for(CNode **i = &nodes[0]; i<&(nodes[0])+numnodes; i++)
  {
    (*i)->Process(u,this);    
  }
}

void CUniverseCell::Update(void)
{
  for(CNode **i = &nodes[0]; i<&(nodes[0])+numnodes; i++)
  {
    (*i)->Update();    
  }
}

CNode *CUniverseCell::FindNode(RealNum x, RealNum y, RealNum r, CNode *except)
{
  RealNum d;
  RealNum mindistance = r*r;
  CNode *rval = NULL;

  for(unsigned n = NEIGHBOUR(-1,-1); n<=NEIGHBOUR(1,1); n++)
  {
    if (neighbours[n])
    {
      for(CNode **i = &(neighbours[n]->nodes[0]);
        i < &(neighbours[n]->nodes[0])+neighbours[n]->numnodes; i++)
      {
        if ((*i) != except)
        {
          if ((d=(*i)->DistanceSquared(x,y)) < mindistance)
          {
            rval = *i;
            mindistance = d;
          }
        }
      }
    }
  }

  return rval;
}

int CUniverseCell::GetSignalValue(RealNum x, RealNum y, RealNum r, CNode *except)
{
  int rval = 0;
  RealNum r2 = r*r;

  for(unsigned n = NEIGHBOUR(-1,-1); n<=NEIGHBOUR(1,1); n++)
  {
    if (neighbours[n])
    {
      for(CNode **i = &(neighbours[n]->nodes[0]);
        i < &(neighbours[n]->nodes[0])+neighbours[n]->numnodes; i++)
      {
        if ((*i) != except)
        {
          (*i)->GetSignalValue(x,y,r2,rval);
        }
      }
    }
  }

  return rval;
}

void CUniverseCell::Centre(RealNum &x, RealNum &y, unsigned &n)
{
  for(CNode **i = &nodes[0]; i<&(nodes[0])+numnodes; i++)
  {
    x += (*i)->GetXpos();
    y += (*i)->GetYpos();
  }

  n += numnodes;
}

void CUniverseCell::Radius(RealNum x, RealNum y, RealNum &d)
{
  RealNum r;

  for(CNode **i = &nodes[0]; i<&(nodes[0])+numnodes; i++)
  {
    r = (*i)->DistanceSquared(x,y);

    if (r>d*d)
    {
      d = sqrt(r);
    }
  }
}

void CUniverseCell::Draw(void *param,RealNum x, RealNum y, RealNum w, RealNum h,unsigned options)
{
  for(CNode **i = &nodes[0]; i<&(nodes[0])+numnodes; i++)
  {
    (*i)->Draw(param,x,y,w,h,options);
  }
}

void CUniverseCell::DrawPostscript(ofstream &os,RealNum x, RealNum y, RealNum w, RealNum h)
{
  for(CNode **i = &nodes[0]; i<&(nodes[0])+numnodes; i++)
  {
    (*i)->DrawPostscript(os,x,y,w,h);
  }
}

void CUniverseCell::DrawAnim(ofstream &os)
{
  for(CNode **i = &nodes[0]; i<&(nodes[0])+numnodes; i++)
  {
    (*i)->DrawAnim(os);
  }
}

void CUniverseCell::Exclude(RealNum x1, RealNum y1, RealNum x2, RealNum y2)
{
  for(CNode **i = &nodes[0]; i<&(nodes[0])+numnodes;)
  {
    if ((*i)->GetXpos() > x1 && (*i)->GetXpos() < x2 &&
       (*i)->GetYpos() > y1 && (*i)->GetYpos() < y2)
    {
      CNode *todel = *i;
      *i = nodes[--numnodes];
      delete todel;
    }
    else
    {
      i++;
    }
  }
}

CUniverse::CUniverse(RealNum cw, RealNum ch)
  : cellw(cw), cellh(ch), iteration(0)
{
	CUniverseCell::width = cellw;
	CUniverseCell::height = cellh;
}

CUniverse::~CUniverse(void)
{
}

CUniverseCell *CUniverse::GetCell(RealNum x, RealNum y, bool makenew)
{
  int ix = POSTOCELLX(x,cellw);
  int iy = POSTOCELLY(y,cellh);

  map<pair<int,int>,CUniverseCell *>::iterator i;

  if ((i=celllookup.find(pair<int,int>(ix,iy))) != celllookup.end())
  {
    return i->second->Use();
  }
  else if (makenew)
  {
    return AddCell(x,y)->Use();
  }
  else
  {
    return NULL;
  }
}

CUniverseCell *CUniverse::AddCell(RealNum x, RealNum y)
{
  /* Assume that there is no cell already at the location, otherwise this
   * would not have been called */

  int ix = POSTOCELLX(x,cellw);
  int iy = POSTOCELLY(y,cellh);

  map<pair<int,int>,CUniverseCell *>::iterator i;
  i=celllookup.find(pair<int,int>(ix,iy));

  assert(i==celllookup.end());

  CUniverseCell *nc = new CUniverseCell();

  nc->SetPosition(((RealNum)ix) * cellw,((RealNum)iy) * cellh);

  celllist.push_back(nc);
  celllookup[pair<int,int>(ix,iy)] = nc;

  {for(int i=-1; i<=1; i++)
  {
    for(int j=-1; j<=1; j++)
    {
      CUniverseCell *c = GetCell(cellw*ix+cellw*i+cellw/2,cellh*iy+cellh*j+cellh/2,false);

      if (c)
      {
        nc->SetNeighbour(NEIGHBOUR(i,j),c);
        c->SetNeighbour(NEIGHBOUR(-i,-j),nc);
      }
    }
  }}

  return nc;     
}

void CUniverse::RemoveCell(CUniverseCell *c)
{
  int ix = POSTOCELLX(c->XPos()+cellw/2,cellw);
  int iy = POSTOCELLY(c->YPos()+cellh/2,cellh);

  celllookup.erase(pair<int,int>(ix,iy));

  for(int i=-1; i<=1; i++)
  {
    for(int j=-1; j<=1; j++)
    {     
      if (CUniverseCell *nc = c->GetNeighbour(NEIGHBOUR(i,j)))
      {
        bool b = nc->GetNeighbour(NEIGHBOUR(-i,-j)) == c;
        assert(b);
        nc->SetNeighbour(NEIGHBOUR(-i,-j),NULL);
      }
    }
  }
}

void CUniverse::ProcessRemovals(void)
{
#if 1
  for(list<CUniverseCell *>::iterator i = celllist.begin();
    i!=celllist.end();)
  {
    if (!(*i)->GetUsed())
    {
      RemoveCell(*i);
      delete *i;
      i = celllist.erase(i);
    }
    else
    {
      (*i)->UnUse();
      i++;
    }
  }
#endif
#if 0
  // shuffle celllist to see whether order makes a difference.
  for(list<CUniverseCell *>::iterator i = celllist.begin();
    i != celllist.end();)
  {
    if (rand()%2==0)
    {
      celllist.push_front(*i);
      i = celllist.erase(i);
    }
    else
    {
      i++;
    }
  }
#endif
}

void CUniverse::AddNode(CNode *n)
{
  GetCell(n->GetXpos(),n->GetYpos())->AddNode(n);
}

void CUniverse::CreateNode(CNode *n)
{
  creatednodes.push_back(n);
}

// should be called after the process step
void CUniverse::PlaceCreatedNodes(void)
{
  // first, check that no two newly created nodes are overlapping
  // this implementation is inefficient , but okay so long as only a few
  // nodes are created every iteration.
  for(list<CNode *>::iterator i = creatednodes.begin(); i!=creatednodes.end();i++)
  {
    bool overlap = false;
    for(list<CNode *>::iterator j = i; j!=creatednodes.end();)
    {
      if (i != j)
      {
        RealNum sumrad = (*i)->GetRadius() + (*j)->GetRadius();
        sumrad = sumrad*sumrad;
        if ( (*i)->DistanceSquared((*j)->GetXpos(),(*j)->GetYpos()) < sumrad)
        {
          // i overlaps with something, delete the thing it overlaps with
          delete (*j);
          j = creatednodes.erase(j);
          overlap = true;
        }
        else
        {
          j++;
        }
      }
      else
      {
        j++;
      }
    }

    // if nothing overlapped with this node, add it to the universe
    if (!overlap)
      AddNode(*i);
    else
      delete (*i);
  }

  creatednodes.clear();
}

void CUniverse::Move(void)
{
  for(list<CUniverseCell *>::iterator i = celllist.begin(); i!=celllist.end();i++)
  {
    (*i)->Move(this);
  }

  if (iteration%100==0)
    ProcessRemovals();
}

void CUniverse::Forces(void)
{
	{for(list<CUniverseCell *>::iterator i = celllist.begin(); i!=celllist.end(); i++)
    (*i)->Forces();
	}

	{for(list<CUniverseCell *>::iterator i = celllist.begin(); i!=celllist.end(); i++)
    (*i)->FrictionForces();
	}
}

void CUniverse::Process(void)
{
  for(list<CUniverseCell *>::iterator i = celllist.begin(); i!=celllist.end(); i++)
    (*i)->Process(this);

  PlaceCreatedNodes();
}

void CUniverse::Update(void)
{
  for(list<CUniverseCell *>::iterator i = celllist.begin(); i!=celllist.end(); i++)
    (*i)->Update();

  iteration++;
}

CNode *CUniverse::FindNode(RealNum x, RealNum y, RealNum r, CNode *except)
{
  CUniverseCell *c = GetCell(x,y);
  if (c)
    return c->FindNode(x,y,r,except);
  else
    return NULL;
}

int CUniverse::GetSignalValue(RealNum x, RealNum y, RealNum r, CNode *except)
{
  CUniverseCell *c = GetCell(x,y);
  if (c)
    return c->GetSignalValue(x,y,r,except);
  else
    return 0;
}

void CUniverse::Centre(RealNum &x, RealNum &y)
{
  unsigned n = 0;
  x = 0;
  y = 0;

  for(list<CUniverseCell *>::iterator i = celllist.begin(); i!=celllist.end(); i++)
    (*i)->Centre(x,y,n);

  x = x / n;
  y = y / n;
}

RealNum CUniverse::Radius(RealNum x, RealNum y)
{
  RealNum d = 0;

  for(list<CUniverseCell *>::iterator i = celllist.begin(); i!=celllist.end(); i++)
    (*i)->Radius(x,y,d);

  return d;
}

void CUniverse::Draw(void *param,RealNum x, RealNum y, RealNum w, RealNum h, unsigned options)
{
  for(list<CUniverseCell *>::iterator i = celllist.begin(); i!=celllist.end(); i++)
    (*i)->Draw(param,x,y,w,h,options);
}

void CUniverse::DrawPostscript(ofstream &os,RealNum x, RealNum y, RealNum w, RealNum h)
{
  for(list<CUniverseCell *>::iterator i = celllist.begin(); i!=celllist.end(); i++)
    (*i)->DrawPostscript(os,x,y,w,h);
}

void CUniverse::DrawAnim(ofstream &os)
{
  for(list<CUniverseCell *>::iterator i = celllist.begin(); i!=celllist.end(); i++)
    (*i)->DrawAnim(os);
}

void CUniverse::Exclude(RealNum x1, RealNum y1, RealNum x2, RealNum y2)
{
  for(list<CUniverseCell *>::iterator i = celllist.begin(); i!=celllist.end(); i++)
    (*i)->Exclude(x1,y1,x2,y2);
}

