// Nodes : node.h
//
// Header file for CUniverse class
//
// See userguide.pdf for copyright information

#ifndef UNIVERSE_H
#define UNIVERSE_H

#include <list>
#include <map>
#include <fstream>

using namespace std;

#include "realnum.h"
#include "node.h"

#define MAX_NODES_PER_CELL 12

class CUniverseCell
{
  public:
    CUniverseCell(void);
    void SetPosition(RealNum x, RealNum y);
    void SetNeighbour(unsigned n, CUniverseCell *uc);
    CUniverseCell *GetNeighbour(unsigned n) {return neighbours[n];}

    void AddNode(CNode *n);
    void Move(CUniverse *u);
    void Forces(void);
    void FrictionForces(void);
    void Process(CUniverse *u);
    void Update(void);
    CNode *FindNode(RealNum x, RealNum y, RealNum r, CNode *except);
    int GetSignalValue(RealNum x, RealNum y, RealNum r, CNode *except);

    void Centre(RealNum &x, RealNum &y, unsigned &n);
    void Radius(RealNum x, RealNum y, RealNum &d);
    void Draw(void *param,RealNum x, RealNum y, RealNum w, RealNum h,unsigned options);
    void DrawPostscript(ofstream &os,RealNum x, RealNum y, RealNum w, RealNum h);
    void DrawAnim(ofstream &os);

    void Dump(ofstream &os);
    RealNum XPos(void) {return xpos;}
    RealNum YPos(void) {return ypos;}

    CUniverseCell *Use(void) {accessed = true; return this;}
    void UnUse(void) {accessed = false;}
    bool GetUsed(void) {return accessed;}

    void Exclude(RealNum x1, RealNum y1, RealNum x2, RealNum y2);
  private:
    RealNum xpos,ypos;

    CUniverseCell *neighbours[9];

    CNode *nodes[MAX_NODES_PER_CELL];
    unsigned numnodes;
    bool accessed;
  public:
    static RealNum width,height;

};

class CUniverse
{
 public:
   ~CUniverse();
  CUniverse(RealNum cw, RealNum ch);

  CUniverseCell *GetCell(RealNum x, RealNum y, bool makenew = true);
  CUniverseCell *AddCell(RealNum x, RealNum y);
  void RemoveCell(CUniverseCell *c);
  void ProcessRemovals(void);

  void AddNode(CNode *n);
  void CreateNode(CNode *n);
  void PlaceCreatedNodes(void);

  void Move(void);
  void Forces(void);
  void Process(void);
  void Update(void);

  CNode *FindNode(RealNum x, RealNum y, RealNum r, CNode *except);
  int GetSignalValue(RealNum x, RealNum y, RealNum r, CNode *except);

  void Centre(RealNum &x, RealNum &y);
  RealNum Radius(RealNum x, RealNum y);

  unsigned GetIteration(void) {return iteration;}

  void Draw(void *param,RealNum x, RealNum y, RealNum w, RealNum h,unsigned options);
  void DrawPostscript(ofstream &os,RealNum x, RealNum y, RealNum w, RealNum h);
  void DrawAnim(ofstream &os);

  void Dump(const char *n);

  void Exclude(RealNum x1, RealNum y1, RealNum x2, RealNum y2);
 private:
   unsigned iteration;
   RealNum cellw,cellh;

   list<CUniverseCell *> celllist;
   list<CNode *> creatednodes;
   map<pair<int,int>,CUniverseCell *> celllookup;
};

#endif
